@@ -1,6 +1,5 @@
 from array import array
 from dis import *
-from new import code, function
 from types import CodeType
 from peak.util.symbols import Symbol
 from peak.util.decorators import decorate_assignment, decorate
@@ -39,6 +38,48 @@ CO_FUTURE_WITH_STATEMENT  = 0x8000      # Python 2.5+ only
 __all__.extend([k for k in globals().keys() if k.startswith('CO_')])
 
 
+
+to_code = lambda x: x.tostring()
+
+try:
+    from new import code as NEW_CODE, function
+except ImportError:
+    from types import FunctionType as function
+    NEW_CODE = lambda ac, *args: CodeType(ac, 0, *args)
+    long = ord = int
+    unicode = basestring = str
+    def to_code(x):
+        global to_code
+        if not hasattr(x, 'tobytes'):
+            to_code = lambda x: x.tostring()
+        else:
+            to_code = lambda x: x.tostring()
+        return to_code(x)
+
+    CODE, GLOBALS, DEFAULTS, CLOSURE, FUNC  = (
+        '__code__', '__globals__', '__defaults__', '__closure__', '__func__'
+    )
+else:
+    CODE, GLOBALS, DEFAULTS, CLOSURE, FUNC  = (
+        'func_code', 'func_globals', 'func_defaults', 'func_closure', 'im_func'
+    )
+    ord = ord
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
 class Const(object):
     """Wrapper to ensure constants are hashable even if mutable"""
 
@@ -83,6 +124,7 @@ class Const(object):
 class Node(tuple):
     """Base class for AST nodes"""
     __slots__ = []
+    __hash__ = tuple.__hash__
 
 def nodetype(*mixins, **kw):
 
@@ -120,7 +162,6 @@ def nodetype(*mixins, **kw):
     return decorate_assignment(callback)
 
 
-
 nodetype()
 def Global(name, code=None):
     if code is None:
@@ -209,19 +250,19 @@ def TryExcept(body, handlers, else_=Pass, code=None):
         return body, tuple(handlers), else_
     okay = Label()
     done = Label()
-    code(
-        okay.SETUP_EXCEPT,
-        body,
-        okay.POP_BLOCK
-    )
+    code(okay.SETUP_EXCEPT, body, okay.POP_BLOCK)
+    if 'POP_EXCEPT' in opcode:
+        code.stack_size += 3
     for typ, handler in handlers:
         next_test = Label()
         Compare(Code.DUP_TOP, [('exception match', typ)], code)
         code(
-            next_test.JUMP_IF_FALSE, Code.POP_TOP,      # remove condition
+            next_test.JUMP_IF_FALSE_OR_POP,             # remove condition
             Code.POP_TOP, Code.POP_TOP, Code.POP_TOP,   # remove exc info
-            handler
         )
+        if 'POP_EXCEPT' in opcode:
+            code(Code.POP_EXCEPT)
+        code(handler)
         if code.stack_size is not None:
             code(done.JUMP_FORWARD)
         code(next_test, Code.POP_TOP)            # remove condition
@@ -302,7 +343,7 @@ def ListComp(body, code=None):
     finally:
         code._tmp_level -= 1
     return r
-    
+
 
 nodetype()
 def LCAppend(value, code=None):
@@ -313,6 +354,9 @@ def LCAppend(value, code=None):
     if sys.version<"2.4":
         code.CALL_FUNCTION(1)
         code.POP_TOP()
+    elif sys.version>="2.7":    # ick
+        code.LIST_APPEND(1)
+        code.POP_TOP()
     else:
         code.LIST_APPEND()
     return r
@@ -323,16 +367,13 @@ def LCAppend(value, code=None):
 
 
 
-
-
-
 nodetype()
 def If(cond, then, else_=Pass, code=None):
     if code is None:
         return cond, then, else_
     else_clause = Label()
     end_if = Label()
-    code(cond, else_clause.JUMP_IF_FALSE, Code.POP_TOP, then)
+    code(cond, else_clause.JUMP_IF_FALSE_OR_POP, then)
     if code.stack_size is not None:
         end_if.JUMP_FORWARD(code)
     code(else_clause, Code.POP_TOP, else_, end_if)
@@ -383,8 +424,7 @@ def Compare(expr, ops, code=None):
         code.DUP_TOP()
         code.ROT_THREE()
         code.COMPARE_OP(op)
-        fail.JUMP_IF_FALSE(code)
-        code.POP_TOP()
+        fail.JUMP_IF_FALSE_OR_POP(code)
     op, arg = ops[-1]
     code(arg)
     code.COMPARE_OP(op)
@@ -408,6 +448,7 @@ deref_to_deref = dict([(k,k) for k in hasfree])
 
 
 
+
 nodetype()
 def And(values, code=None):
     if code is None:
@@ -418,7 +459,7 @@ def And(values, code=None):
             if const_value(value):
                 continue        # true constants can be skipped
         except NotAConstant:    # but non-constants require code
-            code(value, end.JUMP_IF_FALSE, Code.POP_TOP)
+            code(value, end.JUMP_IF_FALSE_OR_POP)
         else:       # and false constants end the chain right away
             return code(value, end)
     code(values[-1], end)
@@ -433,7 +474,7 @@ def Or(values, code=None):
             if not const_value(value):
                 continue        # false constants can be skipped
         except NotAConstant:    # but non-constants require code
-            code(value, end.JUMP_IF_TRUE, Code.POP_TOP)
+            code(value, end.JUMP_IF_TRUE_OR_POP)
         else:       # and true constants end the chain right away
             return code(value, end)
     code(values[-1], end)
@@ -444,10 +485,12 @@ def with_name(f, name):
         return f
     except (TypeError,AttributeError):
         return function(
-            f.func_code, f.func_globals, name, f.func_defaults, f.func_closure
+            getattr(f,CODE), getattr(f,GLOBALS), name, getattr(f,DEFAULTS),
+            getattr(f,CLOSURE)
         )
 
 
+EXTRA_JUMPS = 'JUMP_IF_FALSE_OR_POP JUMP_IF_TRUE_OR_POP JUMP_IF_FALSE JUMP_IF_TRUE'.split()
 
 class Label(object):
     """A forward-referenceable location in a ``Code`` object"""
@@ -470,15 +513,15 @@ class Label(object):
     def POP_BLOCK(self, code):
         self.backpatches[0] = code.POP_BLOCK()
 
-    for op in hasjrel+hasjabs:
-        if opname[op] not in locals():
-            def do_jump(self, code, op=op):
-                method = getattr(code, opname[op])
+    for name in [opname[op] for op in hasjrel+hasjabs]+EXTRA_JUMPS:
+        if name not in locals():
+            def do_jump(self, code, name=name):
+                method = getattr(code, name)
                 if self.resolution is None:
                     return self.backpatches.append(method())
                 else:
                     return method(self.resolution)
-            locals()[opname[op]] = with_name(do_jump, opname[op])
+            locals()[name] = with_name(do_jump, name)
     del do_jump
 
     def __call__(self, code):
@@ -488,8 +531,6 @@ class Label(object):
         for p in self.backpatches:
             if p: p()
 
-
-
 class Code(object):
     co_argcount = 0
     co_stacksize = 0
@@ -654,6 +695,36 @@ class Code(object):
         return len(self.co_code)
 
 
+    if 'UNARY_CONVERT' not in opcode:
+        def UNARY_CONVERT(self):
+            self(Const(repr))
+            self.ROT_TWO()
+            self.CALL_FUNCTION(1, 0)
+
+    if 'BINARY_DIVIDE' not in opcode:
+        def BINARY_DIVIDE(self):
+            self.BINARY_TRUE_DIVIDE()
+
+    if 'DUP_TOPX' not in opcode:
+        def DUP_TOPX(self, count):
+            self.stackchange((count,count*2))
+            if count==2:
+                self.emit(DUP_TOP_TWO)
+            else:
+                raise RuntimeError("Python 3 only supports DUP_TOP_TWO")
+
+    if 'SLICE_0' not in opcode:
+        def SLICE_0(self):
+            self(None, None, Code.SLICE_3)
+        def SLICE_1(self):
+            self(None, Code.SLICE_3)
+        def SLICE_2(self):
+            self(None, Code.ROT_TWO, Code.SLICE_3)
+        def SLICE_3(self):
+            self.BUILD_SLICE(2)
+            self.BINARY_SUBSCR()           
+            
+
     def set_stack_size(self, size):
         if size<0:
             raise AssertionError("Stack underflow")
@@ -664,37 +735,23 @@ class Code(object):
             self.stack_history.extend([self._ss]*bytes)
         self._ss = size
 
+
     def get_stack_size(self):
         return self._ss
 
     stack_size = property(get_stack_size, set_stack_size)
 
-    def stackchange(self, (inputs,outputs)):
+    def stackchange(self, inout):
+        (inputs,outputs) = inout
         if self._ss is None:
             raise AssertionError("Unknown stack size at this location")
         self.stack_size -= inputs   # check underflow
         self.stack_size += outputs  # update maximum height
 
+
     def stack_unknown(self):
         self._ss = None
 
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
     def branch_stack(self, location, expected):
         if location >= len(self.stack_history):
             if location > len(self.co_code):
@@ -715,22 +772,6 @@ class Code(object):
                 "Stack level mismatch: actual=%s expected=%s"
                 % (actual, expected)
             )
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
 
 
 
@@ -741,7 +782,8 @@ class Code(object):
             target = offset
             if op not in hasjabs:
                 target = target - (posn+3)
-                assert target>=0, "Relative jumps can't go backwards"
+                if target<0:
+                    raise AssertionError("Relative jumps can't go backwards")
                 if target>0xFFFF:
                     target = offset - (posn+6)
             return target
@@ -758,12 +800,12 @@ class Code(object):
             self.stack_size += 2
         else:
             old_level = self.stack_size
+        self.stack_size -= (op in (JUMP_IF_TRUE_OR_POP, JUMP_IF_FALSE_OR_POP))
         posn = self.here()
-
         if arg is not None:
             self.emit_arg(op, jump_target(arg))
             self.branch_stack(arg, old_level)
-            lbl = None            
+            lbl = None
         else:
             self.emit_arg(op, 0)
             def lbl(code=None):
@@ -776,7 +818,6 @@ class Code(object):
         self.stackchange((2,1))
         self.emit_arg(COMPARE_OP, compares[op])
 
-
     def setup_block(self, op):
         jmp = self.jump(op)
         self.blocks.append((op,self.stack_size,jmp))
@@ -818,6 +859,47 @@ class Code(object):
             return fwd
 
 
+    if 'JUMP_IF_TRUE_OR_POP' not in opcode:
+        def JUMP_IF_TRUE_OR_POP(self, address=None):
+            lbl = self.JUMP_IF_TRUE(address)
+            self.POP_TOP()
+            return lbl
+        globals()['JUMP_IF_TRUE_OR_POP'] = -1
+
+    if 'JUMP_IF_FALSE_OR_POP' not in opcode:
+        def JUMP_IF_FALSE_OR_POP(self, address=None):
+            lbl = self.JUMP_IF_FALSE(address)
+            self.POP_TOP()
+            return lbl
+        globals()['JUMP_IF_FALSE_OR_POP'] = -1
+
+    if 'JUMP_IF_TRUE' not in opcode:
+        def JUMP_IF_TRUE(self, address=None):
+            self.DUP_TOP()
+            return self.POP_JUMP_IF_TRUE(address)
+    else:
+        globals()['POP_JUMP_IF_TRUE'] = -1
+
+    if 'JUMP_IF_FALSE' not in opcode:
+        def JUMP_IF_FALSE(self, address=None):
+            self.DUP_TOP()
+            return self.POP_JUMP_IF_FALSE(address)
+    else:
+        globals()['POP_JUMP_IF_FALSE'] = -1
+
+    if 'LIST_APPEND' in opcode and LIST_APPEND>=HAVE_ARGUMENT:
+        def LIST_APPEND(self, depth):
+            self.stackchange((depth+1, depth))
+            self.emit_arg(LIST_APPEND, depth)
+
+
+
+
+
+
+
+
+
     def assert_loop(self):
         for why,level,fwd in self.blocks:
             if why==SETUP_LOOP:
@@ -839,7 +921,7 @@ class Code(object):
     def __call__(self, *args):
         last = None
         for ob in args:
-            if callable(ob):
+            if hasattr(ob, '__call__'):
                 last = ob(self)
             else:
                 try:
@@ -855,7 +937,7 @@ class Code(object):
 
     decorate(classmethod)
     def from_function(cls, function, copy_lineno=False):
-        code = cls.from_code(function.func_code, copy_lineno)
+        code = cls.from_code(getattr(function, CODE), copy_lineno)
         return code
 
 
@@ -913,7 +995,7 @@ class Code(object):
 
     def nested(self, name='<lambda>', args=(), var=None, kw=None, cls=None):
         if cls is None:
-            cls = Code
+            cls = self.__class__
         code = cls.from_spec(name, args, var, kw)
         code.co_filename=self.co_filename
         return code
@@ -929,7 +1011,7 @@ class Code(object):
                 oparg = code[i+1] + code[i+2]*256 + extended_arg
                 extended_arg = 0
                 if op == EXTENDED_ARG:
-                    extended_arg = oparg*65536
+                    extended_arg = oparg*long(65536)
                     i+=3
                     continue
                 yield i, op, oparg
@@ -937,7 +1019,7 @@ class Code(object):
             else:
                 yield i, op, None
                 i += 1
-        
+
 
 
 
@@ -1013,13 +1095,13 @@ class Code(object):
         elif parent is not None and self.co_freevars:
             parent.makecells(self.co_freevars)
 
-        return code(
+        return NEW_CODE(
             self.co_argcount, len(self.co_varnames),
-            self.co_stacksize, flags, self.co_code.tostring(),
+            self.co_stacksize, flags, to_code(self.co_code),
             tuple(self.co_consts), tuple(self.co_names),
             tuple(self.co_varnames),
             self.co_filename, self.co_name, self.co_firstlineno,
-            self.co_lnotab.tostring(), self.co_freevars, self.co_cellvars
+            to_code(self.co_lnotab), self.co_freevars, self.co_cellvars
         )
 
 
@@ -1187,14 +1269,95 @@ def fold_args(f, *args):
 
 
 
+def iter_code(codestring):
+    """Iterate over a code string, yielding (start,op,arg,jump,end) tuples
+
+    `start` is the position of the operation start, `end` is the position of
+    the next operation start.  `jump` is a jump target or ``None`` if `op`
+    isn't a jump.  `op` is the opcode, and `arg` the argument, with 32-bit
+    ``EXTENDED_ARG`` instructions pre-processed.
+    """
+    start = ptr = 0
+    size = len(codestring)
+    extend = 0
+    while ptr < size:
+        op = ord(codestring[ptr])
+        ptr += 1
+        if op>=HAVE_ARGUMENT:
+            arg = ord(codestring[ptr]) + ord(codestring[ptr+1])*256 + extend
+            extended_arg = 0
+            ptr += 2
+            if op == EXTENDED_ARG:
+                extend = arg*long(65536)
+                continue
+            if op in hasjrel or op in hasjabs:
+                jump = arg+ptr*(op in hasjrel)
+            else:
+                jump = None
+        else:
+            arg = label = jump = None
+        yield start, op, arg, jump, ptr
+        start = ptr
+
+argtype = {}
+for name, group in dict(
+    co_consts = hasconst,
+    co_names = hasname,
+    co_varnames = haslocal,
+    free = hasfree,
+    cmp_ops = hascompare,
+).items():
+    for op in group:
+        argtype[op] = name
+
+def dump(code):
+    """Disassemble code in a symbolic manner, i.e., without offsets"""
+    code = getattr(code, FUNC, code)
+    code = getattr(code, CODE, code)
+    co_names = code.co_names
+    co_consts = [repr(x) for x in code.co_consts]
+    co_varnames = code.co_varnames
+    cmp_ops = cmp_op
+    free = code.co_cellvars + code.co_freevars
+    labels = {}
+    instructions = list(iter_code(code.co_code))
+    lbl = [jump for start, op, arg, jump, end in instructions if jump is not None]
+    lbl.sort()
+    for jump in lbl:
+        labels.setdefault(jump, "L%d:" % (len(labels)+1))
+
+    i = 0
+    while i<len(instructions):
+        start, op, arg, jump, end = instructions[i]
+        ln = '        '+labels.get(start, '').ljust(7)
+        if op==DUP_TOP and instructions[i+1][1] in (POP_JUMP_IF_FALSE, POP_JUMP_IF_TRUE):
+            s, op, arg, jump, end = instructions[i+1]
+            ln+=' '+['JUMP_IF_FALSE', 'JUMP_IF_TRUE'][op==POP_JUMP_IF_TRUE].ljust(15)
+            i+=1
+        elif op in (JUMP_IF_TRUE_OR_POP, JUMP_IF_FALSE_OR_POP):
+            ln += ' ' + opname[op][:-7].ljust(15)
+        else:
+            ln += ' ' + opname[op].ljust(15)
+        if jump is not None:
+            ln += ' ' + labels[jump][:-1].rjust(10)
+        elif arg is not None:
+            ln += ' ' + repr(arg).rjust(10)
+            if op in argtype:
+                ln += ' (%s)' % (locals()[argtype[op]][arg])
+        print(ln)
+        if op in (JUMP_IF_TRUE_OR_POP, JUMP_IF_FALSE_OR_POP):
+            print('        '+''.ljust(7) + ' POP_TOP')
+        i+=1
+
+
+
 class _se:
     """Quick way of defining static stack effects of opcodes"""
-    POP_TOP   = END_FINALLY = 1,0
+    POP_TOP   = END_FINALLY = POP_JUMP_IF_FALSE = POP_JUMP_IF_TRUE = 1,0
     ROT_TWO   = 2,2
     ROT_THREE = 3,3
     ROT_FOUR  = 4,4
     DUP_TOP   = 1,2
-
     UNARY_POSITIVE = UNARY_NEGATIVE = UNARY_NOT = UNARY_CONVERT = \
         UNARY_INVERT = GET_ITER = LOAD_ATTR = IMPORT_FROM = 1,1
 
@@ -1214,8 +1377,8 @@ class _se:
         (2,0),(3,0),(3,0),(4,0)
     DELETE_SLICE_0, DELETE_SLICE_1, DELETE_SLICE_2, DELETE_SLICE_3 = \
         (1,0),(2,0),(2,0),(3,0)
-
-    STORE_SUBSCR = 3,0
+    
+    STORE_SUBSCR = POP_EXCEPT = EXEC_STMT = 3,0
     DELETE_SUBSCR = STORE_ATTR = 2,0
     DELETE_ATTR = STORE_DEREF = 1,0
     PRINT_EXPR = PRINT_ITEM = PRINT_NEWLINE_TO = IMPORT_STAR = 1,0
@@ -1225,8 +1388,9 @@ class _se:
     LOAD_LOCALS = LOAD_CONST = LOAD_NAME = LOAD_GLOBAL = LOAD_FAST = \
         LOAD_CLOSURE = LOAD_DEREF = IMPORT_NAME = BUILD_MAP = 0,1
 
-    EXEC_STMT = BUILD_CLASS = 3,0
-    JUMP_IF_TRUE = JUMP_IF_FALSE = 1,1
+    BUILD_CLASS = 3,1
+    JUMP_IF_TRUE = JUMP_IF_FALSE = \
+        JUMP_IF_TRUE_OR_POP = JUMP_IF_FALSE_OR_POP = 1,1
 
 if sys.version>="2.5":
     _se.YIELD_VALUE = 1, 1
